/*-
 * Copyright (c) 2013 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

/* a struct for checking randomness */
typedef struct Randomness {
	union {
		uint32_t	*i32;	/* access as 32-bit ints */
		uint8_t		*i8;	/* access as unsigned chars */
	} un;
	uint32_t	bytes;		/* length of input in bytes */
	uint8_t		verbose;	/* gobby? */
} Randomness;

/* test the spread of randomness across nbits */
static int
test_spread(Randomness *b, unsigned nbits, unsigned buckets, int crit)
{
	uint64_t	 hi;
	uint64_t	 num;
	uint32_t	 observed[256];
	int32_t		 chiv[256];
	unsigned	 expected;
	unsigned	 i;
	time_t		 t;
	char		 tim[26];
	char		*s;

	memset(observed, 0x0, sizeof(observed));
	hi = ((uint64_t)1 << nbits);
	for (i = 0 ; i < b->bytes ; i += (nbits / 8)) {
		num = 0;
		memcpy(&num, &b->un.i8[i], nbits / 8);
		observed[num / (hi / buckets)] += 1;
	}
	expected = (b->bytes / (nbits / 8)) / buckets;
	for (i = 0 ; i < buckets ; i++) {
		chiv[i] = (observed[i] - expected);
		chiv[i] = (chiv[i] * chiv[i]) / expected;
		if (chiv[i] > crit) {
			t = time(NULL);
			if ((s = ctime(&t)) == NULL) {
				snprintf(tim, sizeof(tim), "[Time %" PRIi64 " secs]", (int64_t)t);
				s = tim;
			}
			printf("WARNING: %.24s, %u bits, Bucket %d, observed %d, expected %u, chi^2 %d\n",
				s, nbits, i, observed[i], expected, chiv[i]);
		}
	}
	return 1;
}

/* fill the block with n bytes of randomness (from arc4random) */
static int
fill_block(Randomness *b, unsigned bytes)
{
	unsigned	i;

	if (b->bytes == 0) {
		b->bytes = bytes;
		if ((b->un.i8 = calloc(1, b->bytes)) == NULL) {
			return 0;
		}
	}
	for (i = 0 ; i < bytes / sizeof(uint32_t) ; i++) {
		b->un.i32[i] = arc4random();
		if (b->verbose > 2) {
			printf("0x%x ", b->un.i32[i]);
		}
	}
	if (b->verbose) {
		printf("\n");
	}
	return 1;
}

int
main(int argc, char **argv)
{
	Randomness	b;
	unsigned	rounds;
	unsigned	bytes;
	unsigned	secs;
	int		i;

	memset(&b, 0x0, sizeof(b));
	bytes = 256;
	rounds = 1;
	secs = 0;
	while ((i = getopt(argc, argv, "b:r:vw:")) != -1) {
		switch(i) {
		case 'b':
			bytes = strtoul(optarg, NULL, 10);
			break;
		case 'r':
			rounds = strtoul(optarg, NULL, 10);
			break;
		case 'v':
			b.verbose += 1;
			break;
		case 'w':
			secs = strtoul(optarg, NULL, 10);
			break;
		default:
			break;
		}
	}
	for (i = 0 ; i < (int)rounds ; i++) {
		if (!fill_block(&b, bytes)) {
			exit(EXIT_FAILURE);
		}
		test_spread(&b, 8, 8, 3);
		test_spread(&b, 32, 8, 3);
		sleep(secs);
	}
	exit(EXIT_SUCCESS);
}
